#include <inc/mmu.h>
#include <kern/pmap.h>
#include <kern/trap.h>
#include <kern/env.h>
#include <kern/sched.h>
#include <kern/console.h>
#include <kern/printf.h>
#include <kern/picirq.h>
#include <kern/kclock.h>
#include <kern/syscall.h>

#include "bochs.h"

u_int page_fault_mode = PFM_NONE;
static struct Taskstate ts;

/* Interrupt descriptor table.  (Must be built at run time because
 * shifted function addresses can't be represented in relocation records.)
 */
struct Gatedesc idt[256] = { {0}, };
struct Pseudodesc idt_pd =
{
	0, sizeof(idt) - 1, (unsigned long) idt,
};

void
idt_init(void)
{
	extern struct Segdesc gdt[];

	// Setup a TSS so that we get the right stack
	// when we trap to the kernel.
	ts.ts_esp0 = KSTACKTOP;
	ts.ts_ss0 = GD_KD;

	// Love to put this code in the initialization of gdt, but
	// the compiler generates an error incorrectly.
	gdt[GD_TSS >> 3] = SEG16(STS_T32A, (u_long) (&ts),
					sizeof(struct Taskstate), 0);
	gdt[GD_TSS >> 3].sd_s = 0;

	// Set up the handlers: This can be done in a loop since we know the size of handlerX
	SETGATE(idt[0], 0, GD_KT, handler0, 0);
	SETGATE(idt[1], 0, GD_KT, handler1, 0);
	SETGATE(idt[2], 0, GD_KT, handler2, 0);
	SETGATE(idt[3], 0, GD_KT, handler3, 3);
	SETGATE(idt[4], 0, GD_KT, handler4, 0);
	SETGATE(idt[5], 0, GD_KT, handler5, 0);
	SETGATE(idt[6], 0, GD_KT, handler6, 0);
	SETGATE(idt[7], 0, GD_KT, handler7, 0);
	SETGATE(idt[8], 0, GD_KT, handler8, 0);
	SETGATE(idt[9], 0, GD_KT, handler9, 0);

	SETGATE(idt[10], 0, GD_KT, handler10, 0);
	SETGATE(idt[11], 0, GD_KT, handler11, 0);
	SETGATE(idt[12], 0, GD_KT, handler12, 0);
	SETGATE(idt[13], 0, GD_KT, handler13, 0);
	SETGATE(idt[14], 0, GD_KT, handler14, 0);
	SETGATE(idt[15], 0, GD_KT, handler15, 0);
	SETGATE(idt[16], 0, GD_KT, handler16, 0);
	SETGATE(idt[17], 0, GD_KT, handler17, 0);
	SETGATE(idt[18], 0, GD_KT, handler18, 0);
	SETGATE(idt[19], 0, GD_KT, handler19, 0);

	SETGATE(idt[20], 0, GD_KT, handler20, 0);
	SETGATE(idt[21], 0, GD_KT, handler21, 0);
	SETGATE(idt[22], 0, GD_KT, handler22, 0);
	SETGATE(idt[23], 0, GD_KT, handler23, 0);
	SETGATE(idt[24], 0, GD_KT, handler24, 0);
	SETGATE(idt[25], 0, GD_KT, handler25, 0);
	SETGATE(idt[26], 0, GD_KT, handler26, 0);
	SETGATE(idt[27], 0, GD_KT, handler27, 0);
	SETGATE(idt[28], 0, GD_KT, handler28, 0);
	SETGATE(idt[29], 0, GD_KT, handler29, 0);

	SETGATE(idt[30], 0, GD_KT, handler30, 0);
	SETGATE(idt[31], 0, GD_KT, handler31, 0);

	// IRQ
	SETGATE(idt[32], 0, GD_KT, handler32, 0);
	SETGATE(idt[33], 0, GD_KT, handler33, 0);
	SETGATE(idt[36], 0, GD_KT, handler36, 0);

	SETGATE(idt[48], 0, GD_KT, handler48, 3);

	// Load the TSS
	ltr(GD_TSS);

	// Load the IDT
	asm volatile("lidt _idt_pd+2");
}


static void
print_trapframe(struct Trapframe *tf)
{
	printf("TRAP frame at %p\n", tf);
	printf("	edi  0x%x\n", tf->tf_edi);
	printf("	esi  0x%x\n", tf->tf_esi);
	printf("	ebp  0x%x\n", tf->tf_ebp);
	printf("	oesp 0x%x\n", tf->tf_oesp);
	printf("	ebx  0x%x\n", tf->tf_ebx);
	printf("	edx  0x%x\n", tf->tf_edx);
	printf("	ecx  0x%x\n", tf->tf_ecx);
	printf("	eax  0x%x\n", tf->tf_eax);
	printf("	es   0x%x\n", tf->tf_es);
	printf("	ds   0x%x\n", tf->tf_ds);
	printf("	trap 0x%x\n", tf->tf_trapno);
	printf("	err  0x%x\n", tf->tf_err);
	printf("	eip  0x%x\n", tf->tf_eip);
	printf("	cs   0x%x\n", tf->tf_cs);
	printf("	flag 0x%x\n", tf->tf_eflags);
	printf("	esp  0x%x\n", tf->tf_esp);
	printf("	ss   0x%x\n", tf->tf_ss);
}

void
trap(struct Trapframe *tf)
{
  // print_trapframe(tf);
	if (tf->tf_trapno == T_PGFLT) {
		page_fault_handler(tf);
		return;
	}
	page_fault_mode = PFM_NONE;
	if (tf->tf_trapno == T_SYSCALL) {
	  u_int result;
	  result = syscall(tf->tf_eax,
			   tf->tf_edx,
			   tf->tf_ecx,
			   tf->tf_ebx,
			   tf->tf_edi,
			   tf->tf_esi);
	  tf->tf_eax = result;
	  return;
	}
	if (tf->tf_trapno == IRQ_OFFSET+0) {
		// irq 0 -- clock interrupt
	        sched_yield();
	}
	if (tf->tf_trapno == IRQ_OFFSET+1) {
		// irq 1 -- keyboard
	        kbd_intr();
		return;
	}
	if (tf->tf_trapno == IRQ_OFFSET+4) {
		serial_intr();
		return;
	}
	if (IRQ_OFFSET <= tf->tf_trapno 
			&& tf->tf_trapno < IRQ_OFFSET+MAX_IRQS) {
		// just ingore spurious interrupts
		printf("spurious interrupt on irq %d\n",
			tf->tf_trapno - IRQ_OFFSET);
		print_trapframe(tf);
		return;
	}

	// the user process or the kernel has a bug.
	print_trapframe(tf);
	env_destroy(curenv);
	panic("unhandled trap");
}


#define XSTK_PUSH(addr, val) (addr) -= 1; *(TRUP(addr)) = (val);

void
page_fault_handler(struct Trapframe *tf)
{
	u_int fault_va;
	u_int *tos;
	u_int cpl;

	// Read processor's CR2 register to find the faulting address
	fault_va = rcr2();
	cpl = (tf->tf_cs) & 0x3;
	if (cpl & 0x3) {
	  // In user mode
	  page_fault_mode = PFM_KILL;
	  if (curenv->env_pgfault_handler == 0) {
	    //  No handler registered...
	    printf("[%08x] user fault va %08x ip %08x\n", curenv->env_id, fault_va, tf->tf_eip);
	    env_destroy(curenv);
	  } else {
	    if ((tf->tf_esp < curenv->env_xstacktop) &&
		(tf->tf_esp > curenv->env_xstacktop-BY2PG)) {
	      //  Already in handler routine of the user...
	      tos = (u_int *) (tf->tf_esp);
	    } else {
	      //  User program had a page fault -- first time
	      tos = (u_int *) (curenv->env_xstacktop);
	    }
	    XSTK_PUSH(tos, 0);
	    XSTK_PUSH(tos, 0);
	    XSTK_PUSH(tos, 0);
	    XSTK_PUSH(tos, 0);
	    XSTK_PUSH(tos, 0);
	    XSTK_PUSH(tos, tf->tf_eip);
	    XSTK_PUSH(tos, tf->tf_eflags);
	    XSTK_PUSH(tos, tf->tf_esp);
	    XSTK_PUSH(tos, tf->tf_err);
	    XSTK_PUSH(tos, fault_va);
	    // This is where we kind of cheat by knowing the KSTACKTOP always is
	    //  where we will return from
	    asm volatile("movl %2, %%eax\n"
			 "movl %0, -8(%%eax)\n"
			 "movl %1, %%ebx\n"
			 "movl %%ebx, -20(%%eax)\n" : :
			 "g" (tos), "g" (curenv->env_pgfault_handler), "i" (KSTACKTOP)
			 : "ebx" , "eax" );
	  }
	} else {
	  // trap happened in kernel mode
	  if (page_fault_mode == PFM_NONE) {
	    //  Ok, panic in kernel
	    panic("Kernel caused page fault in kernel code\n");
	  } else {
	    // User passed us bad value
	    printf("[%08x] PFM_KILL va %08x ip %08x\n", curenv->env_id, fault_va, tf->tf_eip);
	    env_destroy(curenv);
	  }
	}
}
