// implement fork from user space

#include "lib.h"
#include "../kern/bochs.h"

extern void _asm_pgfault_handler(void);

#define debug 0

//
// Custom page fault handler - if faulting page is copy-on-write,
// map in our own private writable copy.
//

void
cow_pgfault(u_int va, u_int err)
{
  u_char *tmp = (u_char*)(UTEXT-BY2PG);	// should be available!;
  Pte pte;
  
  va = va & 0xfffff000; // Get the base address of the page
  if (debug) sys_cputs("pgfault");
  if (!(err & FEC_WR)) {
    sys_cputs("non-writing page fault\n");
    sys_env_destroy(0);
  }
  pte = vpd[PDX(va)];
  if (pte & PTE_P) {
    pte = vpt[VPN(va)];
    if ((pte & PTE_P) && (pte & PTE_COW)) {
      if ((sys_mem_alloc(0, (u_int)tmp, PTE_P|PTE_U|PTE_W)) < 0) {
	sys_cputs("Failed to allocate memory for copy\n");
	sys_env_destroy(0);
      }
      bcopy((u_char *)va, tmp, BY2PG);
      if (sys_mem_map(0, (u_int)tmp, 0, va, (((pte & ~PTE_COW) | PTE_W) & PTE_USER)) < 0) {
	sys_env_destroy(0);
      }
      if (sys_mem_unmap(0, (u_int)tmp) < 0)
	sys_cputs("Failed to unmap in pgfault");
    } 
  } else {
    sys_cputs("Page fault unexpected\n");
    sys_env_destroy(0);
  }
}

//
// Map our virtual page pn (address pn*BY2PG) into the target envid
// at the same virtual address.  if the page is writable or copy-on-write,
// the new mapping must be created copy on write and then our mapping must be
// marked copy on write as well.  (Exercise: why mark ours copy-on-write again if
// it was already copy-on-write?)
// 
static void
duppage(u_int envid, u_int pn)
{
  int r;
  u_int addr = pn*BY2PG;
  Pte pte;
	
  pte = vpt[VPN(addr)];
  if (pte & PTE_P){
    if (pte & PTE_LIBRARY) {
      // Library pages are always copied no-matter what
      pte = pte & PTE_USER;      
      if ((r = sys_mem_map(0, addr, envid, addr, pte)))
	printf("Failing to map 0x%x ro : %e\n", addr, r);
    } else {
      // Otherwise, copy-on-write writeable pages or COW pages
      if ((pte & PTE_W) || (pte & PTE_COW)) {
	pte = (((pte & ~PTE_W) | PTE_COW) & PTE_USER);
	if ((r = sys_mem_map(0, addr, envid, addr, pte)))
	  printf("Failing to map 0x%x rw : %e\n", addr, r);
	if ((r = sys_mem_map(0, addr, 0, addr, pte)))
	  printf("Failing to remap 0x%x rw : %e\n", addr, r);
      } else {
	// Else just copy it
	pte = pte & (PTE_USER);
	if ((r = sys_mem_map(0, addr, envid, addr, pte)))
	  printf("Failing to map 0x%x ro : %e\n", addr, r);
      }
    }
  }
}

//
// User-level fork.  Create a child and then copy our address space
// and page fault handler setup to the child.
//
// Hint: use vpd, vpt, and duppage.
// Hint: remember to fix "env" in the child process!
// 
int
fork(void)
{
  int envid;
  u_int addr, addr2;
  Pte pte;
  char *envname;

  // install a pgfault handler
  set_pgfault_handler(cow_pgfault);

  // allocate a child environment
  envid = sys_env_alloc(env->env_name);
  if (envid < 0)
    return envid; // error code of env_alloc

  if (envid == 0) {
    env = &envs[ENVX(sys_getenvid())];
    return 0;
  } else {
    addr = 0;
    while (addr < (UTOP-PDMAP)) { // don't do this for the exception stack
      pte = vpd[PDX(addr)];
      if (pte & PTE_P) {
	for (addr2 = addr; addr2 < addr+PDMAP; addr2 += BY2PG) {
	  duppage(envid, addr2/BY2PG);
	}
      }
      addr += PDMAP;
    }
    // Get all final pages up to the USTACKTOP, but not including UXSTACKTOP
    for (addr2 = addr; addr2 < addr+(PDMAP-BY2PG); addr2 += BY2PG) {
      duppage(envid, addr2/BY2PG);
    }

    if (sys_mem_alloc(envid, UXSTACKTOP-BY2PG, PTE_P|PTE_W|PTE_U)) {
      printf("Couldn't map a page for the child exception stack\n");
      return -E_NO_MEM;
    }
    if (sys_set_pgfault_handler(envid, (u_int)_asm_pgfault_handler, UXSTACKTOP)) {
      printf("Failed to set child's pgfault handler\n");
      return -E_UNSPECIFIED;
    }
    if (sys_set_env_status(envid, ENV_RUNNABLE)) {
      printf("Failed to mark child as runnable\n");
      return -E_UNSPECIFIED;
    }
    return (envid); 
  }
}

// Challenge!
int
sfork(void)
{
	panic("sfork not implemented");
	return -E_INVAL;
}
