#ifndef _LIB_H
#define _LIB_H

#include <stdio.h>
#include <inc/types.h>
#include <inc/stdarg.h>
#include <inc/error.h>
#include <inc/env.h>
#include <inc/pmap.h>
#include <inc/syscall.h>
#include <inc/args.h>
#include <inc/fs.h>
#include <user/fd.h>
#include <inc/malloc.h>

#define USED(x) (void)(x)

// ipc.c
void	ipc_send(u_int whom, u_int val, u_int srcva, u_int perm);
u_int	ipc_recv(u_int *whom, u_int dstva, u_int *perm);

// printf.c
void	_panic(const char*, int, const char*, ...);
void	printf(const char*, ...);
int	snprintf(char*, int, const char*, ...);
int	vsnprintf(char*, int, const char*, va_list);
void	warn(const char*, ...);
#define assert(x)	\
	do {	if (!(x)) panic("assertion failed: %s", #x); } while (0)
#define	panic(...) _panic(__FILE__, __LINE__, __VA_ARGS__)
int	fprintf(int fd, const char*, ...);

// string.c
void	memcpy(void*, const void*, u_int);
void	bcopy(const void*, void*, u_int);
void	memset(void*, const char, u_int);
void	bzero(void*, u_int);
char*	strcpy(char*, const char*);
char*	strncpy(char*, const char*, size_t n);
int	strlen(const char*);
int	strcmp(const char*, const char*);
int     strncmp(const char*, const char*, size_t);
const char*	strchr(const char*, char);
const char*     strrchr(const char *s, char);
char * strcat(char * dest, const char * src);

// libos.c or entry.S
extern struct Env *env;
extern struct Env envs[NENV];
extern struct Page pages[];
extern caddr_t _bss_end;
void	exit(void);

// fork.c
#define	PTE_LIBRARY	0x400
#define PTE_COW		0x800

void cow_pgfault(u_int va, u_int err);
int	fork(void);
int	sfork(void);	// Challenge!

// wait.c
void	wait(u_int);

// pageref.c
int	pageref(void*);

// pgfault.c
void	set_pgfault_handler(void(*)(u_int va, u_int err));

// console.c
int	iscons(int);
int	opencons(void);

// syscall.c
void	sys_cputs(char*);
// int	sys_env_alloc(void);
int	sys_env_destroy(u_int);
u_int	sys_getenvid(void);
int	sys_ipc_can_send(u_int, u_int, u_int, u_int);
void	sys_ipc_recv(u_int);
int	sys_set_env_status(u_int, u_int);
int	sys_set_pgfault_handler(u_int, u_int, u_int);
void	sys_yield(void);
int	sys_mem_alloc(u_int, u_int, u_int);
int	sys_mem_map(u_int, u_int, u_int, u_int, u_int);
int	sys_mem_unmap(u_int, u_int);
int	sys_set_trapframe(u_int, struct Trapframe*);
void	sys_panic(char*);
int	sys_cgetc(void);

// This must be inlined.  
// Exercise for reader: why?
static inline int
sys_env_alloc(char *envname)
{
	int ret;

	asm volatile("int %1"
		     : "=a" (ret)
		     : "i" (T_SYSCALL),
		     "a" (SYS_env_alloc),
		     "d" (envname)
	);
	return ret;
}

// fsipc.c
int	fsipc_open(const char*, u_int, u_int, struct Fd*);
int	fsipc_map(u_int, u_int, u_int);
int	fsipc_set_size(u_int, u_int);
int	fsipc_close(u_int);
int	fsipc_dirty(u_int, u_int);
int	fsipc_remove(const char*);
int	fsipc_sync(void);
int	fsipc_incref(u_int);

// fd.c
int	close(int fd);
int	read(int fd, void *buf, u_int nbytes);
int	write(int fd, const void *buf, u_int nbytes);
int	seek(int fd, u_int offset);
void	close_all(void);
int	readn(int fd, void *buf, u_int nbytes);
int	dup(int oldfd, int newfd);
int	dup3(int oldfd);
int	fstat(int fd, struct Stat*);
int	stat(const char *path, struct Stat*);

// file.c
int	open(const char *path, int mode);
int	opendir(const char *path, int mode);
int	read_map(int fd, u_int offset, void **blk);
int	remove(const char *path);
int	ftruncate(int fd, u_int size);
int	sync(void);
int     chdir(const char *);
char *  getcwd(char *);
int     mkdir(const char *, int);
int     isfile(const char *);
int     isdir(const char *);


// pipe.c
int	pipe(int[2]);
int	pipeisclosed(int);

// spawn.c
int	spawn(char*, char**);
int	spawnl(char*, char*, ...);

// envargs.c
char *getenv(char *name);
void delenv(char *name);
int setenv(char *name, char *value);

extern char **environ;


/* File open modes */
#define	O_RDONLY	0x0000		/* open for reading only */
#define	O_WRONLY	0x0001		/* open for writing only */
#define	O_RDWR		0x0002		/* open for reading and writing */
#define	O_ACCMODE	0x0003		/* mask for above modes */

#define	O_CREAT		0x0100		/* create if nonexistent */
#define	O_TRUNC		0x0200		/* truncate to zero length */
#define	O_EXCL		0x0400		/* error if already exists */
#define O_APPEND        0x0800          /* open for append         */
#define O_MKDIR		0x1000		/* create directory, not regular file */

#define MAX_ENV_VARS 32
#define ENVBASE (0xd0000000 - PDMAP - 2*BY2PG)

#endif
