#include "lib.h"
#include <stdio.h>

static int debug = 0;

//
// get the next token from string s
// set *p1 to the beginning of the token and
// *p2 just past the token.
// return:
//	0 for end-of-string
//	> for >
//	| for |
//	w for a word
//
// eventually (once we parse the space where the nul will go),
// words get nul-terminated.
#define WHITESPACE " \t\r\n"
#define SYMBOLS "<|>&;()"


static char **localvars = NULL;
static int localvarcnt = 0;
static int keepgoing = 1;

static void
printlocal()
{
  int i;

  for (i=0; i < localvarcnt; i++) {
    fprintf(STDOUT_FILENO, "%s = %s\n", localvars[i], localvars[i] + strlen(localvars[i]) + 1);
  }
}


static void
unsetlocal(char *name)
{
  int i;

  for (i=0; i <localvarcnt; i++) {
    if (!strcmp(name, localvars[i])) {
      free(localvars[i]);
      if (i < --localvarcnt)
	localvars[i] = localvars[localvarcnt];
    }
  }
}

static char *
getlocal(char *name)
{
  int i;

  for (i=0; i <localvarcnt; i++) {
    if (!strcmp(name, localvars[i])) {
      return (localvars[i] + strlen(name) + 1);
    }
  }
  return NULL;
}

static void
setlocal(char *name, char *value)
{
  int namelen, vallen;
  char *ptr;

  if (!value)
    value = "";
  
  if (getlocal(name))
    unsetlocal(name);

  namelen = strlen(name);
  vallen = strlen(value);

  if ((ptr = malloc(namelen + vallen + 2)) == NULL) {
    fprintf(STDERR_FILE, "Can't set variable\n");
    return;
  }

  strcpy(ptr, name);
  strcpy(ptr+namelen+1, value);
  
  if ((localvars = (char **)realloc((caddr_t)localvars, (localvarcnt + 1) * (sizeof (char *)))) == NULL) {
    panic("Can't reallocate variables\n");
  }
  localvars[localvarcnt] = ptr;
  localvarcnt++;
}


static int
_gettoken(char *s, char **p1, char **p2)
{
	int t;

	if (s == 0) {
		if (debug > 1) fprintf(STDERR_FILE, "GETTOKEN NULL\n");
		return 0;
	}

	if (debug > 1) fprintf(STDERR_FILE, "GETTOKEN: %s\n", s);

	*p1 = 0;
	*p2 = 0;

	while(strchr(WHITESPACE, *s))
		*s++ = 0;
	if(*s == 0) {
		if (debug > 1) fprintf(STDERR_FILE, "EOL\n");
		return 0;
	}
	if(strchr(SYMBOLS, *s)){
		t = *s;
		*p1 = s;
		*s++ = 0;
		*p2 = s;
		if (debug > 1) fprintf(STDERR_FILE, "TOK %c\n", t);
		return t;
	}
	*p1 = s;
	while(*s && !strchr(WHITESPACE SYMBOLS, *s))
	  s++;

	*p2 = s;
	if (debug > 1) {
		t = **p2;
		**p2 = 0;
		fprintf(STDERR_FILE, "WORD: %s\n", *p1);
		**p2 = t;
	}

	return 'w';
}

static char varbuf[1024];

int
gettoken(char *s, char **p1)
{
	static int c, nc;
	static char *np1, *np2;
	char *varptr;

	if (s) {
		nc = _gettoken(s, &np1, &np2);
		return 0;
	}
	c = nc;
	*p1 = np1;
	nc = _gettoken(np2, &np1, &np2);

	if ((*p1) && (**p1 == '$')) {
	  // Variable substitution
	  if ((varptr = getenv(*p1+1))) {
	    strcpy(varbuf, varptr);
	    *p1 = varbuf;
	  } else if ((varptr = getlocal(*p1+1))) {
	    strcpy(varbuf, varptr);
	    *p1 = varbuf;
	  } else {
	    // The variable doesn't exist
	    varbuf[0] = '\0';
	    *p1 = varbuf;
	  }
	}
	return c;
}

static int
shell_run(char *argv[])
{
  int r;

  if (argv == NULL)
    return 0;
  if (!strcmp(argv[0], "setenv")) {
    char *name, *value;
    name = argv[1];
    if (debug) fprintf(STDERR_FILE, "Setting %s\n", argv[1]);
    if (!name) {
      fprintf(STDERR_FILE, "No variable provided -- do you mean printenv?\n");
      return 1;
    }
    if ((value = (char *)strchr(name, '=')) == 0) {
      value = "";
    } else {
      *value++ = '\0';
    }
    setenv(name, value);
  } else if (!strcmp(argv[0], "unsetenv")) {
    if (debug) fprintf(STDERR_FILE, "unsetting %s\n", argv[1]);
    if (argv[1]) {
      delenv(argv[1]);
    }
  } else if (!strcmp(argv[0], "printenv")) {
    int argc = 0;
    if (debug) fprintf(STDERR_FILE, "PRINTENV :\n");    
    while(environ[argc]) {
      fprintf(STDOUT_FILE, environ[argc++]);
      fprintf(STDOUT_FILE, "\n");
    }
  } else if (!strcmp(argv[0], "cd")) {
    if (debug) fprintf(STDERR_FILE, "doing a cd %s\n", argv[1]);    
    if ((r = chdir(argv[1])) < 0)
      fprintf(STDOUT_FILE, "Failing to cd to %s: %e\n", argv[1], r);
  } else if (!strcmp(argv[0], "pwd")) {
    if (debug) fprintf(STDERR_FILE, "pwd\n");
    printf("%s\n", getcwd(NULL));
  } else if (!strcmp(argv[0], "set")) {
    if (debug) fprintf(STDERR_FILE, "setting variable\n");
    if (!argv[1]) {
      printlocal();
      return 1;
    }
    if (!argv[2])
      argv[2] = "";
    setlocal(argv[1], argv[2]);
    return 1;
  } else if (!strcmp(argv[0], "unset")) {
    if (debug) fprintf(STDERR_FILE, "unsetting variable\n");
    if (!argv[1]) {
      fprintf(STDERR_FILE, "usage:  unset varname\n");
      return 0;
    }
    unsetlocal(argv[1]);
  } else if (!strcmp(argv[0], "exit")) {
    if (debug) fprintf(STDERR_FILE, "exitting\n");
    keepgoing = 0;
    return 1;
  } else {
    return 0;
  }
  return 1;
}


#define MAXARGS 32
void
runcmd(char *s)
{
	char *argv[MAXARGS+1], *t;
	int argc, c, i, r, p[2], fd, rightpipe, pid;
	int more, parencnt, have_parencmd;
	int saved_in, saved_out, saved_err;
	char parencmd[1024];
	char *bufptr;

	rightpipe = 0;
	parencnt = 0;
	gettoken(s, 0);
	saved_in = dup3(STDIN_FILENO);
	saved_out = dup3(STDOUT_FILENO);
	saved_err = dup3(STDOUT_FILENO);
again:
	argc = 0;
	more = 0;
	have_parencmd = 0;
	for(;;){
		c = gettoken(0, &t);
		switch(c){
		case 0:
			goto runit;
		case 'w':
		  if (have_parencmd) {
		    fprintf(STDERR_FILE, "Bad ()\'s\n");
		    exit();
		  }
		  if(argc == MAXARGS){
		    fprintf(STDERR_FILE,"too many arguments\n");
		    exit();
		  }
		  argv[argc++] = t;
		  break;
		case ')':
		  /* should not be here unless it's a dangling paren */
		  fprintf(STDERR_FILE, "Dangling close paren\n");
		  exit();
		case '(':
		  if (argc != 0) {
		    fprintf(STDERR_FILE, "Badly placed ()\'s\n");
		    exit();
		  }
		  parencmd[0] = '\0';
		  bufptr = parencmd;
		  argv[0] = "sh";
		  argv[1] = "-c";
		  argv[2] = parencmd;
		  argv[3] = 0;
		  argc = MAXARGS;
		  parencnt++;
		  while (parencnt) {
		    c = gettoken(0, &t);
		    if (c == 'w') {
		      strcpy(bufptr, t);
		      bufptr += strlen(bufptr);
		      *bufptr++ = ' ';
		      *bufptr = '\0';		      
		    } else {
		      if (c == '(')
			parencnt++;
		      if (c == ')')
			parencnt--;
		      if (parencnt != 0) {
			*bufptr++ = c;
			*bufptr = '\0';
		      }
		    }
		  }
		  have_parencmd = 1;
		  break;
		case ';':
		  more = 1;
		  goto runit;
		case '&':
		  if ((pid = fork()) < 0) {
		    fprintf(STDERR_FILE, "Can't create child process\n");
		    exit();
		  }
		  if (pid != 0) {
		    //  Ok lose lose with this layout..
		    //  either we use dup to handle a<b &
		    // or we dont, to handle a &; b.. former is
		    // more likely...
		    dup(saved_in, STDIN_FILE);
		    dup(saved_out, STDOUT_FILE);
		    dup(saved_err, STDERR_FILE);
		    goto again;
		  } else {
		    keepgoing = 0;
		    goto runit;
		  }
		  break; 
		case '<':
			if(gettoken(0, &t) != 'w'){
				fprintf(STDERR_FILE, "syntax error: < not followed by word\n");
				exit();
			}
			if ((fd = open(t, O_RDONLY)) < 0) {
			  fprintf(STDERR_FILE, "Can't open %s for reading\n", t);
			  exit();
			}
			dup(fd, 0);
			close(fd);
			break;
		case '>':
		  if(gettoken(0, &t) != 'w'){
		    fprintf(STDERR_FILE, "syntax error: > not followed by word\n");
		    exit();
		  }
		  // if ((fd = open(t, O_WRONLY|O_CREAT|O_EXCL)) < 0) {
		  if ((fd = open(t, O_WRONLY|O_CREAT|O_TRUNC)) < 0) {
		    fprintf(STDERR_FILE, "Can't open %s for writing: %e (%d)\n", t, fd, fd);
		    exit();
		  }
		  dup(fd, 1);
		  close(fd);
		  break;
		case '|':
			// Your code here.
			// 	First, allocate a pipe.
			//	Then fork.
			//	the child runs the right side of the pipe:
			//		dup the read end of the pipe onto 0
			//		close the read end of the pipe
			//		close the write end of the pipe
			//		goto again, to parse the rest of the command line
			//	the parent runs the left side of the pipe:
			//		dup the write end of the pipe onto 1
			//		close the write end of the pipe
			//		close the read end of the pipe
			//		set "rightpipe" to the child envid
			//		goto runit, to execute this piece of the pipeline
			//			and then wait for the right side to finish
		  if (pipe(p) < 0) {
		    fprintf(STDERR_FILE, "Failing to open a pipe");
		    exit();
		  }
		  if ((rightpipe = fork()) < 0) {
		    fprintf(STDERR_FILE, "Can't fork for pipe\n");
		    exit();
		  }
		  if (rightpipe == 0) {
		    keepgoing = 0;
		    dup(p[0], 0);
		    close(p[0]);
		    close(p[1]);
		    goto again;
		  } else {
		    dup(p[1], 1);
		    close(p[1]);
		    close(p[0]);
		    goto runit;
		  }
		  break;
		}
	}

runit:
	if(!have_parencmd && (argc == 0)) {
	  if (debug) fprintf(STDERR_FILE, "EMPTY COMMAND\n");
	  return;
	}
	argv[argc] = 0;
	if (!shell_run(argv)) {
	  if (debug) {
	    fprintf(STDERR_FILE, "[%08x] SPAWN:", env->env_id);
	    if (have_parencmd)
	      fprintf(STDERR_FILE, " subshell: %s", parencmd);
	    else
	      for (i=0; argv[i]; i++)
		fprintf(STDERR_FILE, " %s", argv[i]);
	    fprintf(STDERR_FILE, "\n");
	  }
	  if ((r = spawn(argv[0], argv)) < 0)
	    fprintf(STDERR_FILE, "spawn %s: %e\n", argv[0], r);
	  if (r >= 0) {
	    if (debug) fprintf(STDERR_FILE, "[%08x] WAIT %s %08x\n", env->env_id, argv[0], r);
	    wait(r);
	    if (debug) fprintf(STDERR_FILE, "[%08x] wait (%s %08x) finished\n", env->env_id, argv[0], r);
	  }
	}
	if (rightpipe) {
	  dup(saved_out, STDOUT_FILE);
	  if (debug) fprintf(STDERR_FILE, "[%08x] WAIT right-pipe %08x\n", env->env_id, rightpipe);
	  wait(rightpipe);
	  if (debug) fprintf(STDERR_FILE, "[%08x] wait (right %08x) finished\n", env->env_id, rightpipe);
	}
	dup(saved_in, STDIN_FILE);
	dup(saved_out, STDOUT_FILE);
	dup(saved_err, STDERR_FILE);
	if (more) {
	  goto again;
	}
	close(saved_in);
	close(saved_out);
	close(saved_err);
}

void
readline(char *buf, u_int n)
{
	int i, r;

	r = 0;
	for(i=0; i<n; i++){
		if((r = read(0, buf+i, 1)) != 1){
			if(r < 0)
			  fprintf(STDERR_FILE, "read error: %e", r);
			exit();
		}
		if(buf[i] == '\b'){
			if(i > 0)
				i -= 2;
			else
				i = 0;
		}
		if(buf[i] == '\n'){
			buf[i] = 0;
			return;
		}
	}
	fprintf(STDERR_FILE, "line too long\n");
	while((r = read(0, buf, 1)) == 1 && buf[0] != '\n')
		;
	buf[0] = 0;
}	

char buf[1024];

void
usage(void)
{
  fprintf(STDERR_FILE, "usage: sh [-dix] [command-file | -c commandstring]\n");
  exit();
}

void
umain(int argc, char **argv)
{
	int r, interactive, echocmds, command;

	interactive = '?';
	echocmds = 0;
	command = 0;

	ARGBEGIN{
	case 'c':
		command = 1;
		break;
	case 'd':
		debug++;
		break;
	case 'i':
		interactive = 1;
		break;
	case 'x':
		echocmds = 1;
		break;
	default:
		usage();
	}ARGEND
	   
	   if (command) {
	     buf[0] = '\0';
	     while (argc-- > 0) {
	       strcat(buf, argv[argc]);
	       strcat(buf, " ");
	     }
	     if (echocmds)
	       fprintf(1, "# %s\n", buf);
	     runcmd(buf);
	     exit();
	   } else {
	     if(argc > 1)
	       usage();
	     if(argc == 1){
	       close(0);
	       if ((r = open(argv[0], O_RDONLY)) < 0)
		 panic("open %s: %e", r);
	       assert(r==0);
	     }
	     if(interactive == '?')
	       interactive = iscons(0);
	   }
	while(keepgoing){
		if (interactive)
			fprintf(1, "$ ");
		readline(buf, sizeof buf);
		if (debug) fprintf(STDERR_FILE, "LINE: %s\n", buf);
		if (buf[0] == '#')
			continue;
		if (echocmds)
			fprintf(1, "# %s\n", buf);
		runcmd(buf);
	}
}

